# -*- coding: utf-8; mode: python -*-
"""A low-level module to read, manipulate and save PAM configuration items"""

import os
import os.path
import errno
import sys

try:
    os.walk
except AttributeError:
    from compat.dummy_os import walk
    os.walk = walk

# peer imports
from support import safeConfigWriter, testset

try: set = set
except NameError: from compat import dummy_set as set

try: import logging
except ImportError: import compat.dummy_logging as logging

theLogger = logging.getLogger("PAMConfig.confset")

def incompleteLine(line):
    """Checks if a given line ends with odd number of backslashes"""
    isIncomplete = False
    for i in xrange(len(line) - 1, -1, -1):
        if line[i] == '\\':
            isIncomplete = not isIncomplete
        else:
            break
    # endfor
    return isIncomplete
# enddef incompleteLine

class PAMFormatError(ValueError):
    """raised when a non-fatal configuration parse error occurs"""
    pass

def PAMNotEnoughTokensError():
    return PAMFormatError("not enough configuration tokens in atom")

class PAMFormatFatalError(ValueError):
    """raised when a fatal configuration parse error occurs"""
    pass

class PAMServiceError(IndexError):
    """raised when a requested service doesn't exist
    (do we need to load \"other\" in this case?)"""
    pass

class PAMControlFlag(object):
    """This class represents control-flag of PAM configuration atom
    (of type TYPE_CONF, see below)
    Typically control-flag is a single word, one of

    * requisite
    * required
    * sufficient
    * optional

    FreeBSD (OpenPAM?) guys also added

    * binding
    
    Unfortunately some ... hot heads in Linux PAM development team
    introduced new brain-da^W complicated syntax of the following
    form:

    [value1=action1 value2=action2 ...]

    where value is a module result code and can be one of the following:

    * success
    * open_err
    * symbol_err
    * service_err
    * system_err
    * buf_err
    * perm_denied
    * auth_err
    * cred_insufficient
    * authinfo_unavail
    * user_unknown
    * maxtries
    * new_authtok_reqd
    * acct_expired
    * session_err
    * cred_unavail
    * cred_expired
    * cred_err
    * no_module_data
    * conv_err
    * authtok_err
    * authtok_recover_err
    * authtok_lock_busy
    * authtok_disable_aging
    * try_again
    * ignore
    * abort
    * authtok_expired
    * module_unknown
    * bad_item
    * default

    default is used when a module returns a code not mentioned
    explicitly in the control-flag entry.

    action can be either a positive number or one of the following
    constants:

    * ignore
    * bad
    * die
    * ok
    * done
    * reset
    """

    # Here's mapping between old and new flags
    # FIXME: Actually need to add 'binding' description
    # of FreeBSD PAM
    PATTERNS = {'requisite': {'success'          : 'ok',
                              'new_authtok_reqd' : 'ok',
                              'ignore'           : 'ignore',
                              'default'          : 'bad'},
                'required':  {'success'          : 'ok',
                              'new_authtok_reqd' : 'ok',
                              'ignore'           : 'ignore',
                              'default'          : 'die'},
                'sufficient':{'success'          : 'done',
                              'new_authtok_reqd' : 'done',
                              'default'          : 'ignore'},
                'optional'  :{'success'          : 'ok',
                              'new_authtok_reqd' : 'ok',
                              'default'          : 'ignore'}
                }

    VALID_VALUES = (
        'success',
        'open_err',
        'symbol_err',
        'service_err',
        'system_err',
        'buf_err',
        'perm_denied',
        'auth_err',
        'cred_insufficient',
        'authinfo_unavail',
        'user_unknown',
        'maxtries',
        'new_authtok_reqd',
        'acct_expired',
        'session_err',
        'cred_unavail',
        'cred_expired',
        'cred_err',
        'no_module_data',
        'conv_err',
        'authtok_err',
        'authtok_recover_err',
        'authtok_lock_busy',
        'authtok_disable_aging',
        'try_again',
        'ignore',
        'abort',
        'authtok_expired',
        'module_unknown',
        'bad_item',
        'default',
        )

    VALID_ACTIONS = (
        'ok',
        'done',
        'bad',
        'die',
        'reset')

    def __init__(self, flag):
        """creates a PAMControlFlag instance based on given flag.
        flag can be either a string or a extended control dictionary or another
        PAMControlFlag instance
        """
        if isinstance(flag, PAMControlFlag):
            self.simple = False
            self._flagset = flag._flagset.copy()
            self._checkSimplicity()
            return

        if isinstance(flag, dict):
            self.simple = False
            self._flagset = flag.copy()
            self._checkSimplicity()
            return

        # the rest is for string
        flag = flag.lower()
        if flag.startswith('['):
            assert flag.endswith(']'), "syntax error in flag specification"
            try:
                # we have a string of the form
                # [v1=a1 v2=a2....]
                # to parse it we first drop braces,
                # then split everything by spaces
                # then each element is splitted by first '='
                # and resulting list of pairs vX, aX are passed
                # to dictionary, which automatically controls
                # that the second split ('=') actually occured
                # (otherwise ValueError is raised)
                values = dict(map(lambda t: t.split('=', 1),
                                  flag[1:-1].split()))
            except ValueError:
                raise PAMFormatError("Invalid control flag: '%s'" % flag)
            self._flagset = values
        else:
            if self.PATTERNS.has_key(flag):
                self._flagset = self.PATTERNS[flag].copy()
            else:
                # emmm... likely a non standard flag
                self._flag = flag
                self.simple = True
                self._flagset = {}
                return
        self._checkSimplicity()
    # enddef __init__

    def copy(self):
        """returns a copy of this control flag"""
        c = self.__class__(self._flagset)
        c._flag = self._flag
        c.simple = self.simple
    # enddef copy

    def __cmp__(self, other):
        if self.simple:
            if isinstance(other, PAMControlFlag):
                return cmp(other, self._flag)
            else:
                return cmp(self._flag, str(other))
        else:
            if isinstance(other, PAMControlFlag):
                return cmp(self._flagset, other._flagset)
            elif self.PATTERNS.has_key(str(other)):
                return cmp(self._flagset, self.PATTERNS[str(other)])
            else:
                return 1
        # endif
    # enddef __cmp__
            
    def __str__(self):
        if self.simple:
            return self._flag
        else:
            aflags = self._flagset.copy()
            # py2.2 dict doesn't have dict.pop...
            # need to decide if we want to fight against
            # this stupid misfeature
            # (below missing of fromkeys is also workarounded)
            try:
                default = aflags['default']
                del aflags['default']
            except KeyError: # there was no default
                default = None
            flist = aflags.items()
            # default goes last
            if default is not None:
                flist.append(('default', default))
            return "[%s]" % (' '.join(map(lambda it: "%s=%s" % it, flist)))
    # enddef __str__

    def update(self, flagset):
        self._flagset.update(flagset)
        self._checkSimplicity()
    # enddef update

    # A single code operation methods
    def get(self, name):
        return self._flagset.get(name, self._flagset.get('default', None))
    # enddef get

    getFlag = get # for backward compatibility

    # Jump getters
    def jumps(self, codes = None):
##        if self.simple:
##            return ()
        results = []
        if not codes:
            codes = self._flagset.iterkeys()
        for k in codes:
            try:
                j = self.jump(k)
            except ValueError:
                pass
            else:
                results.append((k, j))
        # endfor
        return results
    # enddef jumps

    def jump(self, code = 'success'):
        v = self._flagset.get(code, self._flagset.get('default', None))
        if v == 'ok':
            v = 0
        else:
            v = int(v) # raises ValueError upon failure.
        return v
    # enddef jump

    # Jump settters

    def updateJumps(self, delta = 1, filt = lambda p: True, codes = None):
        if not codes:
            codes = self._flagset.iterkeys()
        for c in codes:
            try:
                j = self.jump(c)
            except ValueError:
                pass
            else:
                if filt((c, j)):
                    logging.debug("updateJumps: (j, delta) = (%d, %d)" % (j, delta))
                    j += delta
                    self._setJump(c, j)
                # endif
            # endtry
        # endfor
        self._checkSimplicity()
    # enddef updateJumps

    def setJump(self, code, distance):
        self._setJump(code, distance)
        self._checkSimplicity()
    # enddef setJump
    
    def _setJump(self, code, distance):
        """setJump backend"""
        if distance <= 0:
##            # some damn autoguessing :-/
##            if code in ('ok', 'new_authtok_reqd'):
##                value = 'ok'
##            else:
##                value = 'bad'
            value = 'ok'
        else:
            value = distance
        self._flagset[code] = value
    # enddef _setJump
    
    def _checkSimplicity(self):
        for f,v in self.PATTERNS.iteritems():
            if v == self._flagset:
                self.simple = True
                self._flag = f
                return
        self.simple = False
        self._flag = None
    # enddef _checkSimplicity

# endclass PAMControlFlag

class ConfAtom(object):
    """A class representing a PAM configuration atom
    (typically a single configuration line or comments etc)"""
    # type of lines
    (TYPE_UNPARSED,
     TYPE_INCOMPLETE,
     TYPE_INVALID,
     TYPE_IGNORE,
     TYPE_INCLUDE,
     TYPE_CONF,
     TYPE_CONF_INCLUDE) = range(7) # TYPE_CONF_INCLUDE is for pam_stack.so case

    def __init__(self, isCentral = False, line = None):
        self._unparsed    = []
        if line:
            self._unparsed.append(line)
        self.type    = self.TYPE_UNPARSED

        # Initial ("void") setup
        self.orig        = None
        self.service     = None
        self.facility    = None
        self.control     = None
        self.module      = None
        self.include_fname = None
        self.module_args = []
        self._tokens     = []
        self.comments    = []
        self.isCentral   = isCentral
    # enddef __init__

    def __str__(self):
        if self.orig is not None:
            return self.orig
        elif self.type == self.TYPE_UNPARSED:
            return '' # nothing was added
        elif self.type == self.TYPE_INVALID:
            return '' # well quite unlikely...
        elif self.type == self.TYPE_IGNORE:
            r = ''
        elif self.type == self.TYPE_CONF:
            if self.isCentral:
                assert self.service is not None, "incomplete atom definition!"
                r = self.service + ' '
            else:
                r = ''
            r += ' '.join((self.facility,
                           str(self.control),
                           self.module)) + \
                          ' ' + ' '.join(self.module_args)
        elif self.type == self.TYPE_CONF_INCLUDE:
            r += ' '.join((self.facility,
                           str(self.control),
                           self.module)) + \
                           ' ' + ' '.join(self.module_args)
        elif self.type == self.TYPE_INCLUDE:
            if self.service is not None:
                r = self.service + ' '
            else:
                r = ''
            # we need to handle Debian case specifically
            if self.facility is None:
                r += '@include %s' % self.include_fname
            else:
                r += ' '.join((self.facility,
                               'include',
                               self.include_fname))
        else:
            raise NotImplementedError("Do not know how to coerce %r to string",
                                      (self,))
        if self.comments:
            r += ''.join(self.comments)

        if self.type == self.TYPE_INCOMPLETE:
            endl = '\\\n'
        else:
            endl = '\n'
        
        if not r.endswith(endl):
            r += endl
        return r
    # enddef __str__

    def __repr__(self):
        return "%s.%s(%s)" % (self.__class__.__module__,
                              self.__class__.__name__,
                              repr(str(self)))

    def isSync(self):
        return self.orig is not None

    def append(self, line):
        """appends a line to an internal atom's cache for further
        parsing"""
        assert self.type in (self.TYPE_UNPARSED, self.TYPE_INCOMPLETE), \
               "an attempt to add data to processed atom"
        self._unparsed.append(line)
    # enddef append

    def parse(self):
        """tries to parse all previously added pending lines"""
        if not self._unparsed:
            self.type = self.TYPE_IGNORE
            return
        it = iter(self._unparsed)
        if self.orig is None:
            self.orig = ''
        try:
            for line in it:
                self._parseLine(line)
                self.orig += line
            # endfor
        except PAMFormatError:
            self.type = self.TYPE_INVALID
            # Add unparsed data 
            self.orig += line
            self.orig += ''.join(it)
        # endtry
        self._unparsed = []
    # enddef parse

    def isRequired(self):
        return self.control == "required"

    def isSufficient(self):
        return self.control == "sufficient"

    def controlSet(self, flag):
        self.control = PAMControlFlag(flag)
        self.orig = None
    # enddef controlSet

    def controlUpdate(self, flagset):
        res = self.control.update(flagset)
        self.orig = None
        return res
    # enddef controlUpdate

    def controlUpdateJumps(self, *args, **kwargs):
        res = self.control.updateJumps(*args, **kwargs)
        self.orig = None
        return res
    # enddef controlUpdateJumps

    def _parseLine(self, line):
        # pure-comments line. it's ignored for _all_ cases
        # including those when such a line appears just after
        # a line ending with '\'
        if line.startswith('#'):
            self.comments.append(line)
            if self.type is not self.TYPE_INCOMPLETE:
                self.type = self.TYPE_IGNORE
            return self
        # endif

        # remove comments, including those ending with '\'
        # (checked on Debian, LinuxPAM-0.79 library behaves such a way)
        i = line.find('#')
        if i != -1: # found a comment within the string
            self._tokens.extend(line[:i].split())
        else:
            # check the trailing backslash
            # only in case we didn't encounter a comment
            # note that our line ends with '\n', so we skip it in the check
            if incompleteLine(line[:-1]):
                self._tokens.extend(line[:-2].split())
                # oops, incomplete line
                self.type = self.TYPE_INCOMPLETE
                return self
            else:
                self._tokens.extend(line.split())
            # endif
        # endif

        # the line is complete now and should be parsed

        # self._tokens is a temprary thing so make
        # sure they will be freed upon return:
        tokens = self._tokens
        self._tokens = []
        
        # first skip empty lines (or lines one with whitespaces in)
        if not tokens:
            self.type = self.TYPE_IGNORE
            return self

        # the line(s) is(are) ready for processing.
        # so now we can understand what they mean
        
        # format of tokens is as follows:
        # if this is 'central' pam configuration file
        # (i.e. /etc/pam.conf), self.isCentral is True, then
        # each line should contain the following tokens:
        #
        # <service> <facility> <control> <modulename> [<modulearg> ...]
        #
        # if this is a file from /etc/pam.d/ then <service> is omitted
        # and it's assumed to be the filename:
        #
        # <facility> <control> <modulename> [<modulearg> ...]
        #
        # The first three tokens in case of self.isCentral and the first two
        # otherwise are CASE-INSENSITIVE :-/
        #
        # Inclusion of other configuration files can be done as follows
        # (for /etc/pam.d, not sure about /etc/pam.conf):
        #
        # 1. <facility>   <control>    pam_stack.so service=<filename>  - RedHat9, FC3... / PAM-0.75, PAM-0.77
        # 2. <facility>   include      <filename>                       - SuSE-9.3 / PAM-0.78, ALT Linux, FreeBSD-5.3
        #
        # Debian (as of 3.1) uses the following technique to include
        # other files:
        #
        # 3. @include     <filename>                                    - Debian-3.1 / LinuxPAM-0.79
        #
        # FreeBSD 4.9 and MacOS 10.4 don't seem to use include technique at all
        # Besides FreeBSD 4.9 has only /etc/pam.conf while other systems either
        # don't ship it at all (MacOS 10.4) or simply have it 'just in case'.

        # first check for Debian's '@include' :-/
        if tokens[0] == '@include':
            self.type = self.TYPE_INCLUDE
            self.include_fname = tokens[1]
            return self
        # endif

        # the rest should have 'regular' syntax
        try:
            if self.isCentral:
                self.service = tokens.pop(0).lower()
            # endif
            self.facility = tokens.pop(0).lower()
        except IndexError:
            raise PAMNotEnoughTokensError()

        # the next goes 'control flag'. the problem is that
        # it can actually take several tokens (in case of
        # "new syntax"). So we need to handle them with care:
        (self.control, tokens) = self._fetchControl(tokens)
        
        try:
            self.module = tokens[0]
        except IndexError:
            raise PAMNotEnoughTokensError()
            
        self.module_args = tokens[1:]

        # below this point we actually don't need tokens
        if self.control == 'include' or self.control == 'substack':
            self.type = self.TYPE_INCLUDE
            self.include_fname = self.module
            self.module = None
            self.module_args = None
            return self
        # endif

        if os.path.basename(self.module) == 'pam_stack.so':
            # RedHat case ;-)
            if not testset(lambda arg: arg.startswith('service='), self.module_args):
                raise PAMFormatError("pam_stack.so without corresponding service")

            self.type = self.TYPE_CONF_INCLUDE
            self.include_fname = self.module_args[0][len('service='):]
            return self

        self.type = self.TYPE_CONF
        return self
    # enddef _parseLine

    def _fetchControl(self, tokens):
        if not tokens:
            raise PAMNotEnoughTokensError()
        
        if tokens[0].startswith('['):
            # emm... The new syntax...
            
            # I'm a lazy stupid duck.
            # quite inefficiently but striking
            # (Russian language shines in all its beauty :-))
            rest = ' '.join(tokens)
            try:
                i = rest.index(']') + 1
            except IndexError:
                raise PAMNotEnoughTokensError()
            return (PAMControlFlag(rest[:i]), rest[i:].split())
        else:
            return (PAMControlFlag(tokens.pop(0)), tokens)
    # enddef _fetchControl

    def ConfAtom(cls, facility, control, module, module_args = [], service = None, isCentral = False):
        """A constructor of ConfAtom for TYPE_CONF atom types"""
        # in fact isCentral will be ajusted when the atom will be placed
        # into the ConfAtomList
        obj = cls(isCentral)
        obj.type = obj.TYPE_CONF
        
        obj.service = service # see the comments above
        
        obj.facility = facility
        obj.control = PAMControlFlag(control)
        obj.module = module
        obj.module_args = module_args[:] # a copy. Do we really need to be that cautious?
        return obj
    ConfAtom = classmethod(ConfAtom)
    # enddef ConfAtom

    def CommentAtom(cls, comments = ''):
        # a lame way to handle string arrays as well
        if isinstance(comments, str):
            comments = (comments,)

        processed = list()
        for cline in comments:
            if not cline.endswith('\n'):
                cline += '\n'
            if not cline == '\n' and not cline.startswith('#'):
                cline = '#' + cline
            processed.append(cline)
        # endfor
            
        obj = cls()
        obj.type = obj.TYPE_IGNORE
        obj.comments = processed # replace the old comments completely
        return obj
    CommentAtom = classmethod(CommentAtom)
    # enddef CommentAtom
        
# endclass ConfAtom

class ConfAtomsList(list):
    """Represents a list of atoms (typically read from a PAM configuration file)"""

    def __init__(self, *args):
        list.__init__(self, *args)
        self.service = None
        self.fname = None
        self.isCentral = False
        self._sync = False
        self._loaded = False
    # enddef __init__

    def fromFile(cls, fname, dname =  None):
        """Constructs ConfAtomsList from a given disk file"""
        obj = cls()
        obj._initFromFile(fname, dname)
        return obj
    fromFile = classmethod(fromFile)
    # enddef fromFile

    def ConfAtomsList(cls, fname, dname = None):
        """A constructor to create a PAM service configuration atoms set
        (typically a file) from scratch"""
        obj = cls()
        obj.service = fname
        
        if dname is not None:
            fname = os.path.join(dname, fname)
            obj.isCentral = False
        else:
            obj.isCentral = True

        obj.fname = fname
        return obj
    ConfAtomsList = classmethod(ConfAtomsList)
    # enddef ConfAtomsList
        
    def _initFromFile(self, fname, dname):
        self.service = fname

        if dname is not None:
            fname = os.path.join(dname, fname)
            self.isCentral = False
        else:
            self.isCentral = True
            
        self.fname = fname
        self._loaded = False
        self._sync = False

        f = file(fname)
        
        prevline = None
        for rawline in f.readlines():
            # this should be exception-free
            # otherwise it's a programmer's mistake
            if not prevline:
                prevline = ConfAtom(self.isCentral, rawline)
            else:
                prevline.append(rawline)

            # this can raise a PAMFormatError which should be logged
            # and the processing of the file should be resumed.
            # the problematic line will put as TYPE_INVALID
            try:
                prevline.parse()
            except PAMFormatError:
                logging.exception("ConfAtom parse error:")

            if prevline.type != prevline.TYPE_INCOMPLETE:
                # if this is not a system-wide config, then
                # the service is specified on a file basis,
                # not for each line individually
                if not self.isCentral and prevline.type != prevline.TYPE_IGNORE:
                    prevline.service = self.service
                self.atoms.append(prevline)
                prevline = None
            # endif
        # endfor
        self._loaded = True
        self._sync = True
    # enddef _initFromFile

    def getAtoms(self):
        return self
    atoms = property(getAtoms)

    def getIncludes(self, facility = None):
        """returns a set of all inclusions for a given facility
        (or all if facility is None)
        """
        if facility is None:
            filt = lambda a: a.type in (a.TYPE_INCLUDE, a.TYPE_CONF_INCLUDE)
        else:
            filt = lambda a: a.type in (a.TYPE_INCLUDE, a.TYPE_CONF_INCLUDE) and \
                   (a.facility is None or a.facility == facility)

        return set(map(lambda a: a.include_fname, filter(filt, self.atoms)))
    # enddef getIncludes

    def sync(self):
        """syncronyzes current in-memory state with disk configuration file.
        Old config is safely moved into a backup location"""
        if not self._sync:
            return self._doSync()
        for a in self.atoms:
            if not a.isSync():
                return self._doSync()
        return
    # enddef sync

    def _doSync(self):
        """actual backend for sync."""
        safeConfigWriter(self.fname, map(str, self.atoms), bsuffix = '.saved_by_psa')
    # enddef _doSync

    def __str__(self):
        return ''.join(map(str, self))

    def _bindAtom(self, a):
        """binds the specified atom to itself (usually before including
        it into the atoms list)"""
        a.isCentral = self.isCentral
        if not self.isCentral:
            a.service = self.service
    # enddef _bindAtom

    # below methods to track changes of atoms list in whole
    def __setitem__(self, k, v):
        self._sync = False
        self._bindAtom(v)
        return list.__setitem__(self, k, v)

    def __delitem__(self, k):
        self._sync = False
        return list.__delitem__(self, k)

    def __iadd__(self, other):
        self._sync = False
        map(self._bindAtom, other)
        r = list.__iadd__(self, other)

    def __imul__(self, other):
        raise RuntimeError("__imul__ isn't appropriate for atoms list")

    def append(self, item):
        self._sync = False
        self._bindAtom(item)
        return list.append(self, item)

    def insert(self, i, item):
        self._sync = False
        self._bindAtom(item)
        return list.insert(self, i, item)

    def pop(self, i = -1):
        self._sync = False
        return list.pop(self, i)

    extend = __iadd__ # is this really correct?

    def remove(self, value):
        self._sync = False
        return list.remove(self, value)

    # do we really need these two?
    def reverse(self):
        self._sync = False
        return list.reverse(self)

    def sort(self, cmpfunc = None):
        self._sync = False
        return list.sort(self, cmpfunc)

# endclass ConfAtomsList

class ConfSet(object):
    """Represents a whole set of PAM configurations"""

    def __init__(self, confname = "/etc/pam.conf", confdir = "/etc/pam.d"):
        self.confname = confname
        self.confdir  = confdir
        self.files    = {}
        self.sysconf  = None

        def onerror(err):
            raise err

        fileFilter = lambda fn: not (fn.startswith('.') or fn.lower().find('readme') != -1)

        if confdir:
            dirit = os.walk(confdir, onerror = onerror)
            # do not support subdirs in /etc/pam.d at the moment
            try:
                (dirpath, dirnames, files) = dirit.next()
                # remove those files which start with dot
                if files:
                    files = filter(fileFilter, files)
                else:
                    confdir = None
            except OSError, e:
                if e.errno != errno.ENOENT:
                    raise # re-raise the error
                confdir = None
            # endtry
        # endif

        if confdir:
            for fname in files:
                try:
                    # read errors are treated as fatal
                    self.files[fname] = ConfAtomsList.fromFile(fname, confdir)
                except PAMFormatFatalError:
                    logging.exception('')
                # endtry
            # endfor
        else:
            self.sysconf = ConfAtomsList.fromFile(confname)
            self.files[confname] = self.sysconf
            return
    # enddef __init__

    def getFacilityNames(self):
        """returns PAM facilities names"""
        return ('auth', 'password', 'account', 'session')

    def getServiceNames(self):
        """returns a set of all known services"""
        if self.sysconf:
            # need to collect services names from sysconf
            # filter atoms with TYPE_CONF, retrieve their names
            # and form a set
            return set(map(lambda a: a.service, \
                           filter(lambda a: a.type == a.TYPE_CONF, \
                                  self.sysconf.atoms)))
        else:
            return set(self.files.keys())
        # endif
    # enddef getServiceNames

    def getCommonServiceNames(self, facility = None, services = None):
        """returns a set of common services names.

        A common service is a service which is included from
        another service.

        Optional facility argument allows
        to specify a particular facility to filter services
        """
        if not services:
            services = self.getServiceNames()

        sit = iter(services)

        common = set()
        for s in sit:
            includes = self.getIncludes(s, facility)
            common.update(includes)
        # endfor

        if facility is None:
            return common
        
        # in case facility isn't None
        
        # A special workaround for Debian style of inclusion
        # when 'include' statement doesn't specify facility itself.
        # So we need to go deeper and select only those services
        # which actually has configuration for a given facility
        # This code works for other system as well (but probably
        # suboptimal)
        
        result = set()
        for c in common:
            if self.getConf(c, facility):
                result.add(c)
        return result
    # enddef getCommonServiceNames


    def getConf(self, service, facility):
        """collects and returns a list of configuration atoms
        related for a given facility and service.
        results are returned as a list of 3-element tuples:
    
        0th element of each tuple is the ConfAtom object,
        1st element is the ConfAtomsList object,
        2nd element is an index of the atom in the corresponding ConfAtomsList

        The function attempts to handle inclusions (atoms with type
        TYPE_INCLUDE and TYPE_CONF_INCLUDE) correctly.
        """
        if self.sysconf:
            return self._getConf(self.sysconf, facility, service)
        else:
            if not self.files.has_key(service):
                raise PAMServiceError(service)
            return self._getConf(self.files[service], facility, None)
    # enddef getConf
            
    def _getConf(self, conf, facility, service = None):
        """internal backend for getConf. See getConf documentation"""
        result = []
        
        # try to be nicer for our d^W^W mere mortals 
##        for (atom, i) in zip(conf.atoms, xrange(len(conf.atoms))):
        for i in xrange(len(conf.atoms)):
            atom = conf.atoms[i]
            
            if atom.type == atom.TYPE_CONF and atom.facility == facility and \
               (service is None or atom.service == service):
                result.append((atom, conf, i))
            elif atom.type == atom.TYPE_INCLUDE and \
                 (atom.facility is None or atom.facility == facility) and \
                 (service is None or atom.service == service):
                # need to go one level deeper
                nextsvc = atom.include_fname
                if not self.files or not self.files.has_key(nextsvc):
                    raise PAMServiceError(nextsvc)
                inclist = self._getConf(self.files[nextsvc], facility, None)
                result.extend(inclist)
            elif atom.type == atom.TYPE_CONF_INCLUDE and \
                 (atom.facility == facility) and \
                 (service is None or atom.service == service):
                # need to go one level deeper
                nextsvc = atom.include_fname
                if not self.files or not self.files.has_key(nextsvc):
                    raise PAMServiceError(nextsvc)
                inclist = self._getConf(self.files[nextsvc], facility, None)
                atom.include_size = len(inclist)
                result.append((atom, conf, i))
                result.extend(inclist)
            # endif 
        # endfor
        return result
    # enddef _getConf

    def getIncludes(self, service, facility = None):
        """returns a set of services names directly included into
        a given one (for a particular facility, if specified)
        """
        if self.sysconf:
            # if the whole configuration is in system configuration
            # file, there're no includes at all
            return set()
        else:
            if not self.files.has_key(service):
                raise PAMServiceError(service)

            return self.files[service].getIncludes(facility)
    # enddef getIncludes

    def getAllIncludes(self, service, facility = None):
        """returns a set services names included directly or indirectly
        into a given one (for a particular facility if specified)

        TODO: remove a recursion or at least do it with checks
        """
        incs = self.getIncludes(service, facility)
        for svc in incs.copy():
            incs.update(self.getAllIncludes(svc, facility))
        return incs
    # enddef getAllIncludes

    def createConfAtom(self, service, facility, control, module, module_args = []):
        """OBSOLETED. Use ConfAtom.ConfAtom instead"""
        return ConfAtom.ConfAtom(facility, control, module, module_args, service, bool(self.sysconf))
    # enddef createConfAtom

    def addService(self, service):
        """Adds a service with a given name to this PAM Configuration Set object"""
        if self.sysconf:
            return self.sysconf
        # just in case, sanitize the name
        (dname, fname) = os.path.split(service)
        
        if self.files.has_key(fname):
            return self.files[fname]
        if not self.confdir:
            raise PAMServiceError("separate services aren't supported by this configuration set")
        confList = ConfAtomsList.ConfAtomsList(fname, self.confdir)
        self.files[fname] = confList
        return confList
    # enddef addService

    getService = addService

    def sync(self):
        for f in self.files.values():
            f.sync()
    # enddef sync
# endclass ConfSet

        
